from typing import Any
from .exceptions import MissingPartial, MissingVariable


def missing_variable_default(variable_name: str, variable_tag: str) -> str:
    return ''


def missing_variable_raise(variable_name: str, variable_tag: str):
    raise MissingVariable(variable_name)


def missing_variable_keep(variable_name: str, variable_tag: str) -> str:
    return variable_tag


def missing_partial_default(partial_name: str, partial_tag: str) -> str:
    return ''


def missing_partial_raise(partial_name: str, partial_tag: str) -> str:
    raise MissingPartial(partial_name)


def missing_partial_keep(partial_name: str, partial_tag: str) -> str:
    return partial_tag


def default_serializer(value: Any) -> str:
    """By default, serialize variables as by stringifying"""
    return str(value)



"""Render a Mustache template"""
from typing import Any
from typing import Callable as CallableType
from typing import Dict, List, Optional
from html import escape
from .context import MissingVariable, get_from_context
from .exceptions import MustacheSyntaxError
from .handlers import default_serializer, missing_partial_default, missing_variable_default
from .tokenizer import Token, tokenize
from .types import should_iterate


# pylint:disable=too-many-locals,too-many-branches,too-many-statements,too-many-arguments
def render(
    template: str,
    context: Dict,
    serializer: Optional[CallableType[[Any], str]] = None,
    partials: Optional[Dict] = None,
    missing_variable_handler: Optional[CallableType[[str, str], str]] = None,
    missing_partial_handler: Optional[CallableType[[str, str], str]] = None,
) -> str:
    """Render a mustache template"""

    serializer = serializer or default_serializer
    missing_variable_handler = missing_variable_handler or missing_variable_default
    missing_partial_handler = missing_partial_handler or missing_partial_default

    partials = partials or {}

    output: str = ''
    context_stack: List = [context]
    env_stack: List = []
    left_delimiter: str = '{{'
    right_delimiter: str = '}}'
    pointer: int = 0

    while True:
        try:
            (token, value, indentation), pointer = next(
                tokenize(template, pointer, left_delimiter, right_delimiter)
            )
        except StopIteration:
            break

        current_context = context_stack[-1]

        if token is Token.SET_DELIMITER:
            new_delimiters = value.strip().split(' ')
            left_delimiter = new_delimiters[0]
            right_delimiter = new_delimiters[-1]

        if token is Token.END:
            current_env = env_stack[-1]
            context_stack.pop()

            env_name, env_pointer, [env_var, _] = current_env

            if should_iterate(env_var):
                current_env[2][1] += 1
                try:
                    next_item = env_var[current_env[2][1]]
                    context_stack.append(next_item)
                    pointer = env_pointer
                    continue
                except IndexError:
                    pass

            if env_name != value:
                raise MustacheSyntaxError.from_template_pointer(
                    f'Unexpected section end tag on line {{line_number}}. Expected "{env_name}" got "{value}"',
                    template,
                    pointer,
                )

            env_stack.pop()

        if not current_context and len(context_stack) != 1:
            if token in [Token.SECTION, Token.INVERTED]:
                context_stack.append(False)
                env_stack.append([value, pointer, [False, 0]])
            continue

        if token in [Token.NO_ESCAPE, Token.VARIABLE, Token.SECTION, Token.INVERTED]:
            try:
                variable = get_from_context(context_stack, value)
            except MissingVariable:
                variable = missing_variable_handler(
                    value, f'{left_delimiter} {value} {right_delimiter}'
                )
        else:
            variable = None

        if token is Token.LITERAL:
            output += value

        elif token is Token.NO_ESCAPE:
            output += serializer(variable)

        elif token is Token.VARIABLE:
            output += escape(serializer(variable))

        elif token in [Token.SECTION, Token.INVERTED]:
            if token is Token.INVERTED:
                variable = not variable

            if should_iterate(variable):
                try:
                    context_item = variable[0]
                    context_stack.append(context_item)
                except IndexError:
                    context_stack.append(False)
            else:
                context_stack.append(variable)
            env_stack.append([value, pointer, [variable, 0]])

        elif token is Token.PARTIAL:

            partial_template = partials.get(value)  # potentially raise error here
            if partial_template is None:
                partial_template = missing_partial_handler(
                    value, f'{left_delimiter} {value} {right_delimiter}'
                )

            if partial_template != '':
                remove_trailing_indentation = False
                if partial_template.endswith('\n'):
                    remove_trailing_indentation = True

                partial_template = indentation + f'\n{indentation}'.join(
                    partial_template.split('\n')
                )

                if remove_trailing_indentation:
                    partial_template = partial_template[: -len(indentation)]

                partial_output = render(
                    partial_template, current_context, serializer=serializer, partials=partials
                )
                output += partial_output

    return output



from typing import Any
from collections.abc import Iterator, Sequence


def is_lambda(context: Any) -> bool:
    """Shoule we traeat context value as a lambda?"""
    return callable(context)


def should_iterate(context: Any) -> bool:
    """Should we iterate context value?"""
    return isinstance(context, (Sequence, Iterator)) and not isinstance(context, str)



"""Generate tokens from a mustache template"""
from typing import Iterator, Tuple
from enum import Enum
from .exceptions import MustacheSyntaxError


class Token(Enum):
    """Mustache token type"""

    LITERAL = 'LITERAL'
    VARIABLE = 'VARIABLE'
    COMMENT = '!'
    SECTION = '#'
    INVERTED = '^'
    END = '/'
    PARTIAL = '>'
    SET_DELIMITER = '='
    NO_ESCAPE_BRACE = '{'
    NO_ESCAPE = '&'


def find_next_tag(template: str, pointer: int, left_delimiter: str) -> Tuple[str, int]:
    """Find the next tag, and the literal between current pointer and that tag"""

    split_index = template.find(left_delimiter, pointer)

    if split_index == -1:
        return (template[pointer:], len(template))

    return (template[pointer:split_index], split_index)


def parse_tag(
    template: str, pointer: int, left_delimiter: str, right_delimiter: str
) -> Tuple[Tuple[Token, str], int]:
    """Parse a tag from a template"""
    tag_pointer = pointer + len(left_delimiter)

    try:
        token = Token(template[tag_pointer])
        tag_pointer += 1
    except ValueError:
        token = Token.VARIABLE

    if token is Token.NO_ESCAPE_BRACE:
        right_delimiter = f'}}{right_delimiter}'
        token = Token.NO_ESCAPE
    elif token is Token.SET_DELIMITER:
        right_delimiter = f'={right_delimiter}'

    tag_end_pointer = template.find(right_delimiter, tag_pointer)

    if tag_end_pointer == -1:
        raise MustacheSyntaxError.from_template_pointer(
            'Unclosed tag on line {line_number}', template, pointer
        )

    tag = template[tag_pointer:tag_end_pointer]

    return ((token, tag.strip()), tag_end_pointer + len(right_delimiter))


def find_next_pointer(
    template: str, tag_start_pointer: int, tag_end_pointer: int
) -> Tuple[bool, int, int]:
    """Find the next parsing pointer, based on the current tag and whether it is a standalone"""
    pre_tag_line = template.rfind('\n', 0, tag_start_pointer)
    post_tag_line = template.find('\n', tag_end_pointer)

    indentation_pointer = pre_tag_line
    if pre_tag_line == -1:
        pre_tag_line = 0
        indentation_pointer = 0
    else:
        indentation_pointer += 1

    before_tag = template[pre_tag_line:tag_start_pointer]
    after_tag = template[tag_end_pointer:post_tag_line]
    is_standalone = (before_tag.isspace() or before_tag == '') and (
        after_tag.isspace() or after_tag == ''
    )

    if post_tag_line == -1:
        post_tag_line = len(template)
    else:
        # Skip the newline character
        post_tag_line += 1
    return (
        is_standalone,
        post_tag_line if is_standalone else tag_end_pointer,
        indentation_pointer if is_standalone else tag_start_pointer,
    )


def tokenize(
    template: str, pointer: int = 0, left_delimiter: str = '{{', right_delimiter: str = '}}'
) -> Iterator[Tuple[Tuple[Token, str, str], int]]:
    """Generate tokens from a template string"""

    while pointer < len(template):
        literal, tag_start_pointer = find_next_tag(template, pointer, left_delimiter)
        indentation = ''

        if tag_start_pointer == len(template):
            # Reached the end of the template - return the final literal and end iteration
            yield (Token.LITERAL, literal, ''), len(template)
            break

        (token, token_value), tag_end_pointer = parse_tag(
            template, tag_start_pointer, left_delimiter, right_delimiter
        )

        if token is Token.SET_DELIMITER:
            new_delimiters = token_value.strip().split(' ')
            left_delimiter = new_delimiters[0]
            right_delimiter = new_delimiters[-1]

        if token not in [Token.VARIABLE, Token.NO_ESCAPE]:
            is_standalone, pointer, indentation_pointer = find_next_pointer(
                template, tag_start_pointer, tag_end_pointer
            )

            if is_standalone:
                if tag_start_pointer - indentation_pointer > 0:
                    literal = literal[0 : -(tag_start_pointer - indentation_pointer)]

                if token is Token.PARTIAL:
                    indentation = template[indentation_pointer:tag_start_pointer]

        else:
            # Variable interpolation tags cannot be standalones
            pointer = tag_end_pointer

        if literal != '':
            yield (Token.LITERAL, literal, ''), tag_start_pointer

        # Comments and set delimiters do not appear in output
        if token not in [Token.COMMENT]:
            yield (token, token_value, indentation), pointer



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['moosetash']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'moosetash',
    'version': '0.2.0',
    'description': 'Mustache template renderer',
    'long_description': '# Moosetash\n\n## A Mustache template renderer for Python\n\nMoosetash is a python implementation of the [Mustache](https://mustache.github.io/) templating specification.\n\n## License\n\nThis project is licensed under the MIT License - see the [LICENSE](LICENSE) file for details.\n',
    'author': 'Michael Curtis',
    'author_email': 'michaelrccurtis@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/michaelrccurtis/moosetash',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Utilities for accessing rendering context"""
from typing import Any, List
from .exceptions import ContextAccessError, MissingVariable


def get(obj: Any, variable: str) -> Any:
    """Lookup key from obj"""
    try:
        return obj[variable]
    except (TypeError, AttributeError):
        pass
    except KeyError as ex:
        raise MissingVariable(variable) from ex

    try:
        return getattr(obj, variable)
    except (TypeError, AttributeError):
        pass

    try:
        return obj[int(variable)]
    except (IndexError, KeyError) as ex:
        raise MissingVariable(variable) from ex
    except ValueError:
        pass

    raise MissingVariable(variable)


def deep_get(context: Any, variable: str) -> Any:
    """Recursively fetch a variable from context"""
    if variable == '.':
        # Handle implicit interpolation
        return context

    units = variable.split('.')
    value = context
    has_match = False
    for unit in units:
        try:
            value = get(value, unit)
        except MissingVariable as ex:
            if has_match:
                return None
            raise ex
        except Exception as ex:
            raise ContextAccessError(variable) from ex
        has_match = True
    return value


def get_from_context(contexts: List[Any], variable: str) -> Any:
    """Search list of context for a variable"""

    value = None

    for context in reversed(contexts):
        try:
            value = deep_get(context, variable)
            break
        except MissingVariable:
            pass

    if value is None:
        raise MissingVariable(variable)

    return value



"""Moosetash Exceptions"""


class MustacheSyntaxError(Exception):
    """Error during parsing template"""

    @classmethod
    def from_template_pointer(cls, msg: str, template: str, pointer: int):
        """Create syntax error with positional information"""

        line_number = template[:pointer].count('\n') + 1
        return cls(msg.format(line_number=line_number))


class ContextAccessError(Exception):
    """Error getting a variable from context"""


class MissingVariable(Exception):
    """Variable is missing from context"""


class MissingPartial(Exception):
    """Partial is missing from partials dictionary"""
